from .IOH_function import IOH_function, custom_IOH_function
from .IOH_logger import IOH_logger
from .IOH_Utils import runParallelFunction

from itertools import product
from functools import partial
from multiprocessing import cpu_count

import numpy as np
       
def _run_default(alg, fid, dim, iid, precision, suite, repetitions, observing,
        location, foldername, dat, cdat, idat, tdat_base, tdat_exp,
        parameters, dynamic_attrs, static_attrs):
    '''A helper function for parallellization of the IOHexperimter.
    '''
    name = alg.__class__.__name__
    info = "Run using the IOHexperimenter in python, beta version"
    f = IOH_function(fid, dim, iid, precision, suite)
    if observing:
        logger = IOH_logger(location, foldername, name, info)
        logger.set_tracking_options(dat, cdat, idat, tdat_base, tdat_exp)
        if parameters is not None:
            logger.track_parameters(alg, parameters)
        if dynamic_attrs is not None:
            logger.set_dynamic_attributes(dynamic_attrs)
        if static_attrs is not None:
            logger.set_static_attributes(static_attrs)
        f.add_logger(logger)

    for rep in range(repetitions):
        np.random.seed = rep
        if rep > 0:
            f.reset()
        alg(f)
        print(fid, f.evaluations, f.best_so_far_precision)

    f.clear_logger()
    
def _run_custom(alg, function, fname, dim, suite, repetitions, observing,
        location, foldername, dat, cdat, idat, tdat_base, tdat_exp,
        parameters, dynamic_attrs, static_attrs):
    '''A helper function for parallellization of the IOHexperimter.
    '''
    name = alg.__class__.__name__
    info = "Run using the IOHexperimenter in python, beta version"
    f = custom_IOH_function(function, fname, dim)

    if observing:
        logger = IOH_logger(location, foldername, name, info)
        logger.set_tracking_options(dat, cdat, idat, tdat_base, tdat_exp)
        if parameters is not None:
            logger.track_parameters(alg, parameters)
        if dynamic_attrs is not None:
            logger.set_dynamic_attributes(dynamic_attrs)
        if static_attrs is not None:
            logger.set_static_attributes(static_attrs)
        f.add_logger(logger)

    for rep in range(repetitions):
        np.random.seed = rep
        if rep > 0:
            f.reset()
        alg(f)
        print(f.fid, f.evaluations, f.best_so_far_precision)
    f.clear_logger()
                        
class IOHexperimenter():
    '''An interfact to easily run a set of algorithms on multiple functions from the IOHexperimenter.
    '''
    def __init__(self):
        '''Create an IOHexperimenter object for benchmarking a set of algorithms on multiple functions

        Initialize the functions to use by calling 'initialize_PBO', 'initialize_BBOB' or 'initialize_custom'
        Set up parallellization by calling 'set_parallell'

        '''
        self.location = None
        self.foldername = None
        self.dat = True
        self.cdat = False
        self.idat = 0
        self.tdat_base = [0]
        self.tdat_exp = 0
        self.observing = False
        self.parameters = None
        self.dynamic_attrs = None
        self.static_attrs = None
        self.precision = 0
        self.parallel_settings = None
    
    def initialize_PBO(self, fids, iids, dims, repetitions):
        '''Initialize to a set of functions for the PBO suite

        Parameters
        ----------
        fids:
            List of the function numbers within the selected suite. Alternatively, the names of the functions as used in IOHexperimenter.
        dims:
            List of the dimensions (number of variables) of the problem
        iids:
            List of the instance IDs of the problem
        repetitions: 
            Number of repetitions on each function instance
        '''
        self.fids = fids
        self.iids = iids
        self.dims = dims
        self.repetitions = repetitions
        self.suite = "PBO"
    
    def initialize_BBOB(self, fids, iids, dims, repetitions, precision = 1e-8):
        '''Initialize to a set of functions for the BBOB suite

        Parameters
        ----------
        fids:
            List of the function numbers within the selected suite. Alternatively, the names of the functions as used in IOHexperimenter.
        dims:
            List of the dimensions (number of variables) of the problem
        iids:
            List of the instance IDs of the problem
        repetitions: 
            Number of repetitions on each function instance
        target_precision:
            Optional, how close to the optimum the problem is considered 'solved'
        '''
        self.fids = fids
        self.iids = iids
        self.dims = dims
        self.repetitions = repetitions
        self.precision = precision
        self.suite = "BBOB"
    
    def initialize_custom(self, functions, fnames, fdims, repetitions, suite = "No Suite"):
        '''Initialize to a set of custom functions

        Parameters
        ----------
        functions:
            List of functions to use
        fnames:
            List of names corresponding to the provided functions
        fdims:
            List of dimensionalities corresponding to the provided functions
        repetitions: 
            Number of repetitions on each function instance
        suite:
            Optional, a name for the suite of the provided functions
        '''
        self.functions = functions
        self.fnames = fnames
        self.fdims = fdims
        self.repetitions = repetitions
        self.suite = suite
    
    def set_logger_location(self, location, foldername = "run"):
        '''Set the location options for the logger of the suite
        Parameters
        ----------
        location:
            The directory in which to store the results
        foldername:
            The name of the folder to create to store the results
        
        '''
        self.location = location
        self.foldername = foldername
        self.observing = True
        
    def set_parallel(self, parallel, version = "joblib", timeout = 30, num_threads = None):
        '''Set the parallellization options for the experiments
        
        Parameters
        ----------
        parallel:
            Boolean. Whether or not to use parallell execution
        version:
            Which parallellization library to use. Options are:
                - 'multiprocessing'
                - 'MPI' (using the schwimmbad library)
                - 'joblib' (recommended) 
                - 'pebble' (can timeout functions which take too long)
        num_threads:
            How many threads to use. Defaults to all available ones from 'cpu_count'
        timeout:
            If using pebble, this sets the timeout in seconds after which to cancel the execution
        '''
        self.parallel_settings = {
            "evaluate_parallel" : parallel, "use_MPI" : version == "MPI", 
            "use_pebble" : version == "pebble", "timeout" : timeout, 
            "use_joblib" : version == "joblib",
            "num_threads" : num_threads if num_threads is not None else cpu_count()
        }
        
    def set_logger_options(self, dat = True, cdat = False, idat = 0, tdat_base = [0], tdat_exp = 0):
        '''Set which datafiles should be stored by the logger

        Parameters
        ----------
        dat:
            Whether or not to store .dat-files (one line per improvement of objective)
        cdat:
            Whether or not to store .cdat-files (one line for each evaluation of objective)
        idat:
            Integer, if non-zero, an .idat file will be made which records ever i-th evaluation of the objective
        tdat_base:
            Base values for the tdat-file
        tdat_exp:
            Exponent-values for the tdat-file
        '''
        self.dat = dat
        self.cdat = cdat
        self.idat = idat
        self.tdat_base = tdat_base
        self.tdat_exp = tdat_exp
        
    def set_parameter_tracking(self, parameters):
        '''Set which parameters should be tracked during the benchmarking procedure
        
        Only usable when 'set_logger_location' has been called previously
        
        Parameters
        ----------
        parameters:
            List of the names of the parameters to track. All 'parameters'-variables need to be accessible by 
            using algorithm.parameter (recommended to use @property) for each of the algorithms to be benchmarked
        '''
        self.parameters = parameters
    
    def set_dynamic_tracking(self, attributes):
        '''Set which attributes should be tracked for each algorithm run (stored at the end of the run in the meta-information)
        
        Only usable when 'set_logger_location' has been called previously
        
        Parameters
        ----------
        attributes:
            List of the names of the parameters to track. All 'attributes'-variables need to be accessible by 
            using algorithm.parameter (recommended to use @property) for each of the algorithms to be benchmarked
        '''        
        self.dynamic_attrs = attributes

    def set_static_tracking(self, attributes):
        '''Set which static attributes should be stored in the meta-information of the benchmark data
        
        Only usable when 'set_logger_location' has been called previously
        
        Parameters
        ----------
        attributes:
            List of the names of the parameters to track. All 'attributes'-variables need to be accessible by 
            using algorithm.parameter (recommended to use @property) for each of the algorithms to be benchmarked
        '''        
        self.static_attrs = attributes
    
    def __call__(self, algorithms):
        '''Run the experiment with the given set of algorithms

        Parameters
        ----------
        algorithms:
            A list containing algorithm functions. 
            These will get as input only one argument, of type IOH_function.
        '''
        if self.suite in ["PBO", "BBOB"]:
            arguments = product(algorithms, self.fids, self.dims, self.iids)
            partial_run = partial(_run_default, precision = self.precision, suite = self.suite, 
                      repetitions = self.repetitions, observing = self.observing, 
                      location = self.location, foldername = self.foldername,
                      dat = self.dat, cdat = self.cdat, 
                      idat = self.idat, tdat_base = self.tdat_base, 
                      tdat_exp = self.tdat_exp, parameters = self.parameters,
                      dynamic_attrs = self.dynamic_attrs, static_attrs = self.static_attrs)
        else:
            arguments = product(algorithms, self.functions, self.fnames, self.fdims)
            partial_run = partial(_run_custom, suite = self.suite, 
                      repetitions = self.repetitions, observing = self.observing, 
                      location = self.location, foldername = self.foldername,
                      dat = self.dat, cdat = self.cdat, 
                      idat = self.idat, tdat_base = self.tdat_base, 
                      tdat_exp = self.tdat_exp, parameters = self.parameters,
                      dynamic_attrs = self.dynamic_attrs, static_attrs = self.static_attrs)
            
        results = runParallelFunction(partial_run, arguments, self.parallel_settings)   
        return results


from . import IOHprofiler as IOH
from .IOH_Utils import runParallelFunction
from .IOH_logger import IOH_logger

import pkg_resources, operator
import numpy as np

from pathlib import Path
from functools import partial
from itertools import product
from collections.abc import Iterable

bbob_fid_dict = {
    1: IOH.Sphere, 2: IOH.Ellipsoid, 3: IOH.Rastrigin, 4: IOH.Bueche_Rastrigin, 
    5: IOH.Linear_Slope, 6: IOH.Attractive_Sector, 7: IOH.Step_Ellipsoid, 
    8: IOH.Rosenbrock, 9: IOH.Rosenbrock_Rotated, 10: IOH.Ellipsoid_Rotated,
    11: IOH.Discus, 12: IOH.Bent_Cigar, 13: IOH.Sharp_Ridge, 14: IOH.Different_Powers,
    15: IOH.Rastrigin_Rotated, 16: IOH.Weierstrass, 17: IOH.Schaffers10, 
    18: IOH.Schaffers1000, 19: IOH.Griewank_RosenBrock, 20: IOH.Schwefel,
    21: IOH.Gallagher101, 22: IOH.Gallagher21, 23: IOH.Katsuura, 24: IOH.Lunacek_Bi_Rastrigin
}

pbo_fid_dict = {
    1: IOH.OneMax, 2: IOH.LeadingOnes, 3: IOH.Linear, 4: IOH.OneMax_Dummy1, 
    5: IOH.OneMax_Dummy2, 6: IOH.OneMax_Neutrality, 7: IOH.OneMax_Epistasis, 
    8: IOH.OneMax_Ruggedness1, 9: IOH.OneMax_Ruggedness2, 10: IOH.OneMax_Ruggedness3,
    11: IOH.LeadingOnes_Dummy1, 12: IOH.LeadingOnes_Dummy2, 13: IOH.LeadingOnes_Neutrality,
    14: IOH.LeadingOnes_Epistasis, 15: IOH.LeadingOnes_Ruggedness1, 
    16: IOH.LeadingOnes_Ruggedness2, 17: IOH.LeadingOnes_Ruggedness3, 18: IOH.LABS, 
    19: IOH.MIS, 20: IOH.Ising_Ring, 21: IOH.Ising_Torus, 22: IOH.Ising_Triangular,
    23: IOH.NQueens
}


class IOH_function():
    '''A wrapper around the functions from the IOHexperimenter
    '''

    def __init__(self, fid, dim, iid, target_precision=0, suite="BBOB"):
        '''Instansiate a problem based on its function ID, dimension, instance and suite

        Parameters
        ----------
        fid:
            The function ID of the problem in the suite, or the name of the function as string
        dim:
            The dimension (number of variables) of the problem
        iid:
            The instance ID of the problem
        target_precision:
            Optional, how close to the optimum the problem is considered 'solved'
        suite:
            Which suite the problem is from. Either 'BBOB' or 'PBO'. Only used if fid is an integer

        '''
        if isinstance(fid, int):
            if suite == "BBOB":
                self.f = bbob_fid_dict[fid](iid, dim)
            elif suite == "PBO":
                if fid in [21, 23]:
                    if not np.sqrt(dim).is_integer():
                        raise Exception("For this function, the dimension needs to be a perfect square!")
                self.f = pbo_fid_dict[fid](iid, dim)
            else:
                raise Exception("This suite is not yet supported")
        else:
            if fid in ["Ising_2D", "NQueens"]:
                if not np.sqrt(dim).is_integer():
                    raise Exception("For this function, the dimension needs to be a perfect square!")
            exec (f"self.f = IOH.{fid}({iid}, {dim})")

        self.f.reset_problem()
        self.precision = target_precision
        self.logger = None
        self.track_oob = False
        self.suite = suite
        self.maximization = (self.suite == "PBO")
        self.y_comparison = operator.gt if self.maximization else operator.lt
        self.xopt = None
        self.yopt = (self.maximization - 0.5) * -np.inf

    def __call__(self, x):
        '''Evaluates the function in point x and deals with logging in needed

        Parameters
        ----------
        x:
            The point to evaluate

        Returns
        ------
        The value f(x)
        '''
        y = self.f.evaluate(x)
        if self.y_comparison(y, self.yopt):
            self.yopt = y
            self.xopt = x
        if self.track_oob:
            self.oob += any(x < self.lowerbound) or any(x > self.upperbound)
        if self.logger is not None:
            self.logger.process_parameters()
            #             for param_obj in self.tracked_params:
            #                 print(param_obj)
            #                 exec(f"print({param_obj}.__name__)")
            #                 exec(f"print({param_obj})")
            #                 self.logger.update_parameter(param_obj.__name__, param_obj)
            if self.f.IOHprofiler_get_problem_type() == "bbob":
                self.logger.process_evaluation(self.f.loggerCOCOInfo())
            else:
                self.logger.process_evaluation(self.f.loggerInfo())
        return y

    @property
    def final_target_hit(self):
        '''Returns whether the target has been reached

        Returns
        ------
        A boolean indicating whether the optimal value has been reached within the
        pre-defined precision (default 0)
        '''
        if self.f.IOHprofiler_get_problem_type() == "bbob":
            return self.f.IOHprofiler_get_best_so_far_raw_objectives()[0] <= self.precision
        else:
            return self.f.IOHprofiler_get_best_so_far_raw_objectives()[0] >= self.f.IOHprofiler_get_optimal()[
                0] - self.precision

    @property
    def evaluations(self):
        '''Returns number of evaluations performed so far

        Returns
        ------
        Number of evaluations performed so far
        '''
        return self.f.IOHprofiler_get_evaluations()

    @property
    def best_so_far_precision(self):
        '''Gets the best-so-far raw function value. Do NOT use directly in
        the optimization algorithm, since that would break the black-box assumption!
        instead, use best_so_far_fvalue.

        Returns
        ------
        The raw (precision in case of bbob, untransformed objective in case of PBO) objective value
        '''
        return self.f.IOHprofiler_get_best_so_far_raw_objectives()[0]

    @property
    def best_so_far_variables(self):
        '''Gets the best variables evaluated so far

        Returns
        ------
        The best-so-far variables
        '''
        return self.xopt

    @property
    def best_so_far_fvalue(self):
        '''Gets the best-so-far function value as seen by the algorithm.

        Returns
        ------
        The best-so-far objective value
        '''
        return self.yopt

    @property
    def number_of_variables(self):
        '''Returns the number of variables (dimension) of the problem

        Returns
        ------
        The number of variables (dimension) of the problem
        '''
        return self.f.IOHprofiler_get_number_of_variables()

    @property
    def upperbound(self):
        '''Returns the upper bound of the search space

        Returns
        ------
        A vector containing the upper bound per coordinate
        '''
        return np.array(self.f.IOHprofiler_get_upperbound())

    @property
    def lowerbound(self):
        '''Returns the lower bound of the search space

        Returns
        ------
        A vector containing the lower bound per coordinate
        '''
        return np.array(self.f.IOHprofiler_get_lowerbound())

    @property
    def number_out_of_bounds(self):
        '''Returns the number of times an out-of-bounds solution was evaluated
        This is only tracked is the parameter 'track_oob' is set to True on intiailization of this object

        Returns
        ------
        The number of the number of out-of-bounds solutions evaluated
        '''
        if self.track_oob:
            return self.oob
        else:
            return np.nan

    def add_logger(self, logger):
        '''Adds an observer to the problem

        Parameters
        ----------
        logger:
            A logger object (of class IOH_logger) with which to track this problem.
        '''
        if not isinstance(logger, IOH_logger):
            raise TypeError
        
        logger.track_problem(
            self.f.IOHprofiler_get_problem_id(), self.f.IOHprofiler_get_number_of_variables(),
            self.f.IOHprofiler_get_instance_id(), self.f.IOHprofiler_get_problem_name(),
            self.f.IOHprofiler_get_optimization_type(), self.suite
        )
        self.logger = logger

    def clear_logger(self):
        '''Clears the logger: finishes writing info-files and closes file connections
        '''
        if self.logger is not None:
            self.logger.clear_logger()

    def get_target(self, raw=False):
        '''Returns the target value to reach

         Parameters
        ----------
        raw:
            Whether or not to add the defined precision to the target
        Returns
        ------
        Returns the target value of the function
        '''
        target_raw = self.f.IOHprofiler_get_optimal()[0]
        if raw:
            return target_raw
        return self.precision + target_raw

    def reset(self):
        '''Resets the internal problem. This clears all information about previous evaluations.
        '''
        self.f.reset_problem()
        self.yopt = (self.maximization - 0.5) * -np.inf
        self.xopt = None
        if self.logger is not None:
            self.logger.track_problem(self.f.IOHprofiler_get_problem_id(), self.f.IOHprofiler_get_number_of_variables(),
                                  self.f.IOHprofiler_get_instance_id(), self.f.IOHprofiler_get_problem_name(),
                                  self.f.IOHprofiler_get_optimization_type(), self.suite)

    def enable_oob_tracking(self):
        '''Enables the tracking of number of out-of-bounds points evaluated. Can be accessed trough property 'number_out_of_bounds'
        '''
        self.track_oob = True
        self.oob = 0

class custom_IOH_function(IOH_function):
    '''A wrapper to turn any python function into an IOH_function
    '''

    def __init__(self, internal_eval, fname, dim, fid=0, iid=0, maximization=False, suite="No Suite", upperbound=5,
                 lowerbound=-5):
        '''Convert a regular function into an IOH_function, allowing for easy use with other IOHexperimenter functions, such as IOH_logger

        Parameters
        ----------
        internal_eval:
            The function which will be evaluated. Should take only a vector as its arguments
        fname:
            The name of the problem
        dim:
            The dimension (number of variables) of the problem
        fid:
            A numerical identifier for the problem
        iid:
            The instance ID of the problem
        maximization:
            Boolean indicating whether to maximize (True) or minimize (False). Defaults to False
        suite:
            Optional, indicating which suite of functions this problem is from
        upperbound:
            The upperbound of the problem. Either one value (same for each dimension) or a vector of lenght dim
        lowerbound:
            The upperbound of the problem. Either one value (same for each dimension) or a vector of lenght dim

        '''
        self.internal_eval = internal_eval
        self.maximization = maximization
        self.yopt = (self.maximization - 0.5) * -np.inf
        self.evals = 0
        self.fname = fname
        self.dim = dim
        self.fid = fid
        self.iid = iid
        self.suite = suite
        if isinstance(upperbound, Iterable) and len(upperbound) == dim:
            self.ub = np.array(upperbound)
        else:
            self.ub = np.array(dim * [upperbound])
        if isinstance(lowerbound, Iterable) and len(lowerbound) == dim:
            self.lb = np.array(lowerbound)
        else:
            self.lb = np.array(dim * [lowerbound])
        self.logger = None
        self.y_comparison = operator.gt if self.maximization else operator.lt
        self.track_oob = False

    def __call__(self, x):
        y = self.internal_eval(x)
        self.evals += 1
        if self.y_comparison(y, self.yopt):
            self.xopt = x
            self.yopt = y
        if self.track_oob:
            self.oob += any(x < self.lb) or any(x > self.ub)
        if self.logger is not None:
            self.logger.process_parameters()
            self.logger.process_evaluation((self.evaluations, y, self.yopt, y, self.yopt))
        return y

    @property
    def final_target_hit(self):
        return False

    @property
    def evaluations(self):
        return self.evals

    @property
    def best_so_far_precision(self):
        return self.yopt

    @property
    def number_of_variables(self):
        return self.dim

    @property
    def upperbound(self):
        return self.ub

    @property
    def lowerbound(self):
        return self.lb

    def add_logger(self, logger):
        if not isinstance(logger, IOH_logger):
            raise TypeError
        logger.track_problem(self.fid, self.dim,
                             self.iid, self.fname,
                             self.maximization, self.suite)
        self.logger = logger

    def get_target(self, raw=False):
        return (self.maximization - 0.5) * np.inf

    def reset(self):
        self.yopt = (self.maximization - 0.5) * -np.inf
        self.evals = 0
        self.xopt = None
        if self.logger is not None:
            self.logger.track_problem(self.fid, self.dim, self.iid, self.fname,
                                        self.maximization, self.suite)

class W_model_function(IOH_function):
    '''A wrapper around the W-model functions from the IOHexperimenter.
    '''

    def __init__(self, base_function="OneMax", iid=0, dim=16, dummy=0, epistasis=0,
                 neutrality=0, ruggedness=0):
        '''Instansiate a problem based on its function ID, dimension, instance and suite

        Parameters
        ----------
        base_function:
            The base function on which the W-model should be applied. Currently only OneMax and LeadingOnes are supported
        dim:
            The dimension (number of variables) of the problem
        iid:
            The instance ID of the problem
        dummy:
            Float between 0 and 1, fractoin of valid bits.
        epistasis:
            size of sub-string for epistasis
        neutrality:
            size of sub-string for neutrality
        ruggedness:
            gamma for ruggedness layper
        '''
        if base_function == "OneMax":
            self.f = IOH.W_Model_OneMax(iid, dim)
        elif base_function == "LeadingOnes":
            self.f = IOH.W_Model_LeadingOnes(iid, dim)
        assert epistasis <= dim, "Epistasis has to be less or equal to than dimension"
        assert neutrality <= dim, "Neutrality has to be less than or equal to dimension"
        assert ruggedness <= dim ** 2, "Ruggedness has to be less than or equal to dimension squared"
        assert dummy <= 1 and dummy >= 0, "Dummy variable fraction has to be in [0,1]"
        self.f.set_w_setting(dummy, epistasis, neutrality, ruggedness)
        self.f.reset_problem()
        self.precision = 0
        self.logger = None
        self.track_oob = False
        self.suite = "W_model"
        self.maximization = True
        self.y_comparison = operator.gt
        self.xopt = None
        self.yopt = -np.inf



import os

from .utils import opt_alg, TmpdirTestCase
from IOHexperimenter import IOH_function, IOH_logger


class LoggerTestCase(TmpdirTestCase):    
    def test_initialization_logger(self):
        l = IOH_logger(self.tmpdir, "temp_run")
        l.clear_logger()
        assert 1
        
    def test_logging_fct(self):
        f = IOH_function(1, 5, 1, suite = "BBOB")
        l = IOH_logger(self.tmpdir, "temp_run")
        f.add_logger(l)
        f([0] * 5)
        assert os.path.exists(f"{self.tmpdir}/temp_run")
        f.clear_logger()
        
    def test_logging_with_params(self):
        f = IOH_function(1, 5, 1, suite = "BBOB")
        l = IOH_logger(self.tmpdir, "temp_run")
        a = opt_alg(10000)
        l.track_parameters(a, 'param_rate')
        l.set_static_attributes('static_param')
        l.set_dynamic_attributes('dynamic_prop')
        f.add_logger(l)
        a(f)
        assert os.path.exists(f"{self.tmpdir}/temp_run")
        f.clear_logger()

if __name__ == '__main__':
    unittest.main()


from . import IOHprofiler as IOH

import numpy as np
from pathlib import Path
from collections.abc import Iterable

class IOH_logger:
    def __init__(self, location, foldername = "run", name = "", info = ""):
        '''Creates a logger to be used within an IOH_problem object

        Parameters
        ----------
        location:
            The directory in which to store the results
        foldername:
            The name of the folder to create to store the results
        name:
            The name of the algorithm intened to run on this function
        info:
            Additional information on the algorithm
        
        Returns
        ------
        A logger from the IOHexperimenter package
        '''
        Path(location).mkdir(parents=True, exist_ok=True)
        self.logger = IOH.IOHprofiler_csv_logger(location, foldername, name, info)
        self.logger.activate_logger()
        self.params = []
        self.dynamic_attrs = []
        self.alg = None
        self.set_tracking_options()
    
    def set_tracking_options(self, dat = True, cdat = False, idat = 0, tdat_base = [0], tdat_exp = 0):
        '''Set which datafiles should be stored

        Parameters
        ----------
        dat:
            Whether or not to store .dat-files (one line per improvement of objective)
        cdat:
            Whether or not to store .cdat-files (one line for each evaluation of objective)
        idat:
            Integer, if non-zero, an .idat file will be made which records ever i-th evaluation of the objective
        tdat_base:
            Base values for the tdat-file
        tdat_exp:
            Exponent-values for the tdat-file
        '''
        self.logger.set_complete_flag(cdat) # Set to True to enable cdat
        self.logger.set_interval(idat) # Set to non-zero to enable idat
        self.logger.set_time_points(tdat_base, tdat_exp) # tdat
        self.logger.set_update_flag(dat) # dat
        
    def track_problem(self, problem_id, nr_variables, instance_id, problem_name, optimization_type, suite = "No Suite"):
        '''Set the problem-specific variables to store in the info-file

        Parameters
        ----------
        problem_id:
            Numerical identifier for the problem
        nr_variables:
            How many variables (dimension) the problem has
        instance_id:
            The numerical instance id
        problem_name:
            The name of the problem as a string
        optimization_type:
            Boolean. True for maximization, False for mimimization
        suite:
            Optional. The suite the problem comes from
        '''
        self.logger.track_problem(problem_id, nr_variables,
                              instance_id, problem_name,
                              optimization_type)
        self.logger.track_suite(suite)
        
    def track_parameters(self, algorithm, parameters):
        '''Initialize tracking of parameters during the run of the algorithm

        Parameters
        ----------
        algorithm:
            The algorithm object. Can also be a parameter subobject. The only requirement is that all
            'parameters'-variables are accessible by using algorithm.parameter (recommended to use @property)
        parameters:
            A list of parameters to track
        '''
        if isinstance(parameters, str):
            self.params = [parameters]
        elif isinstance(parameters, Iterable):
            self.params = parameters
        else:
            raise TypeError("parameters needs to be a string or iterable containing strings")
        self.logger.set_parameters_name(self.params)
        self.alg = algorithm
        
    def set_static_attributes(self, attrs):
        '''Set static attributes to store in the info-files. Make sure to only 
        call after setting the tracked algorithm using track_parameters.

        Parameters
        ----------
        attrs:
            A list of parameters to track (should be accessible in the tracked algorithm)
        '''
        if isinstance(attrs, str):
            exec(f"self.logger.add_attribute('{attrs}', self.alg.{attrs})")
        elif isinstance(attrs, Iterable):
            for a in attrs:
                exec(f"self.logger.add_attribute('{a}', self.alg.{a})")
        else:
            raise TypeError("attrs needs to be a string or iterable containing strings")
            
    def set_dynamic_attributes(self, attrs):
        '''Initialize tracking of parameters during the run of the algorithm. Make sure to only 
        call after setting the tracked algorithm using track_parameters.

        Parameters
        ----------
        attrs:
            A list of parameters to track (should be accessible in the tracked algorithm)
        '''
        if isinstance(attrs, str):
            self.dynamic_attrs = [attrs]
        elif isinstance(attrs, Iterable):
            self.dynamic_attrs = attrs
        else:
            raise TypeError("attrs needs to be a string or iterable containing strings")
        self.logger.set_dynamic_attributes_name(self.dynamic_attrs)

    def process_evaluation(self, info):
        '''Internal function for processing evalutation

        Parameters
        ----------
        info:
            The specific information from the function which needs to be processed
        '''
        self.logger.write_line(info)
        
    def process_parameters(self):
        '''Internal function for processing parameter changes

        '''
        for param in self.params:
            exec(f"self.logger.set_parameters(['{param}'], [self.alg.{param}])")
        for param in self.dynamic_attrs:
            exec(f"self.logger.set_dynamic_attributes(['{param}'], [self.alg.{param}])")
    
    def clear_logger(self):
        '''Function to clear the logger once it is no longer needed. Finishes all the writing to file which is still in underlying buffers.

        '''
        self.logger.clear_logger()


from functools import partial
from multiprocessing import cpu_count

def runParallelFunction(runFunction, arguments, parallel_config = None):
    """
        Return the output of runFunction for each set of arguments,
        making use of as much parallelization as possible on this system

        :param runFunction: The function that can be executed in parallel
        :param arguments:   List of tuples, where each tuple are the arguments
                            to pass to the function
        :param force_single:Force the run to be single-threaded, ignoring all settings from IOH_config
        :param timeout:     How many seconds each execution get before being stopped. 
                            Only used when running in parallel without mpi.
        :return:
    """
    if parallel_config is None:
        return runSingleThreaded(runFunction, arguments)    

    if parallel_config['evaluate_parallel']:
        if parallel_config['use_MPI']:
            return runMPI(runFunction, arguments)
        elif parallel_config['use_pebble']:
            return runPebblePool(
                runFunction, arguments, 
                timeout = parallel_config['timeout'], 
                num_threads = parallel_config['num_threads']
            )
        elif parallel_config['use_joblib']:
            return runJoblib(
                runFunction, arguments, 
                num_threads = parallel_config['num_threads']
            )
        else:
            return runPool(
                runFunction, arguments, 
                num_threads = parallel_config['num_threads']
            )
    else:
        return runSingleThreaded(runFunction, arguments)

# Inline function definition to allow the passing of multiple arguments to 'runFunction' through 'Pool.map'
def func_star(a_b, func):
    """Convert `f([1,2])` to `f(1,2)` call."""
    return func(*a_b)

def runPool(runFunction, arguments, num_threads = None):
    """
        Small overhead-function to handle multi-processing using Python's built-in multiprocessing.Pool

        :param runFunction: The (``partial``) function to run in parallel, accepting ``arguments``
        :param arguments:   The arguments to passed distributedly to ``runFunction``
        :return:            List of any results produced by ``runFunction``
    """
    if num_threads is None:
        num_threads = cpu_count()
    arguments = list(arguments)
    print(f"Running pool with {min(num_threads, len(arguments))} threads")
    from multiprocessing import Pool
    p = Pool(min(num_threads, len(arguments)))

    local_func = partial(func_star, func=runFunction)
    results = p.map(local_func, arguments)
    p.close()
    return results

def runPebblePool(runfunction, arguments, timeout = 30, num_threads = None):
    from pebble import ProcessPool
    from concurrent.futures import TimeoutError
    
    if num_threads is None:
        num_threads = cpu_count()
    
    arguments = list(arguments)
    
    print(f"Running pebble pool with {min(num_threads, len(arguments))} threads")

    
    def task_done(future):
        try:
            future.result()  # blocks until results are ready
        except TimeoutError as error:
            print("Function took longer than %d seconds" % error.args[1])
        except Exception as error:
            print("Function raised %s" % error)
            print(error)  # traceback of the function
            
    with ProcessPool(max_workers = min(num_threads, len(arguments)), max_tasks = len(arguments)) as pool:
        for x in arguments:
            if type(x) is not list and type(x) is not tuple:
                x = [x]
            future = pool.schedule(runfunction, args=x, timeout=timeout)
            future.add_done_callback(task_done)

def runJoblib(runFunction, arguments, num_threads = None):
    from joblib import delayed, Parallel
    
    if num_threads is None:
        num_threads = cpu_count()
#     arguments = list(arguments)
#     print(f"Running joblib with {min(num_threads, len(arguments))} threads")
    local_func = partial(func_star, func=runFunction)

    return Parallel(n_jobs = num_threads)(delayed(local_func)(arg) for arg in arguments)

def runSingleThreaded(runFunction, arguments):
    """
        Small overhead-function to iteratively run a function with a pre-determined input arguments

        :param runFunction: The (``partial``) function to run, accepting ``arguments``
        :param arguments:   The arguments to passed to ``runFunction``, one run at a time
        :return:            List of any results produced by ``runFunction``
    """
    results = []
    print("Running single-threaded")
    for arg in arguments:
        results.append(runFunction(*arg))
    return results

def runMPI(runFunction, arguments):
    from schwimmbad import MPIPool
    print("Running using MPI")
    with MPIPool() as pool:
        results = pool.map(runFunction, arguments)
    return results
  



import unittest
import numpy as np

from IOHexperimenter import IOH_function, custom_IOH_function, W_model_function


class PBOTestCase(unittest.TestCase):
    def test_initialization_PBO(self):
        f = IOH_function(1, 16, 1, suite = "PBO")
        assert isinstance(f, IOH_function)
        assert f.maximization
        assert f.number_of_variables == 16
        assert f.suite == "PBO"
        assert not f.final_target_hit
        assert all(f.upperbound == 1)
        assert all(f.lowerbound == 0)
        
    def test_initialization_BBOB(self):
        f = IOH_function(1, 5, 1, suite = "BBOB")
        assert isinstance(f, IOH_function)
        assert not f.maximization
        assert f.number_of_variables == 5
        assert f.suite == "BBOB"
        assert not f.final_target_hit
        assert all(f.upperbound == 5)
        assert all(f.lowerbound == -5)
        
    def test_initialization_custom(self):
        def internal_func(x):
            return np.sum(x)
        f = custom_IOH_function(internal_func, "sum", 5)
        assert isinstance(f, IOH_function)
        assert not f.maximization
        assert f.number_of_variables == 5
        assert f.suite == "No Suite"
        assert not f.final_target_hit
        assert all(f.upperbound == 5)
        assert all(f.lowerbound == -5)

    def test_W_model(self):
        f = W_model_function()
        assert isinstance(f, IOH_function)
        assert f.maximization
        assert f.number_of_variables == 16
        assert f.suite == "W_model"
        assert not f.final_target_hit
        f2 = IOH_function(1, 16, 0, suite = "PBO")
        assert f(16 * [0]) == f2(16 * [0])
        assert f(16 * [1]) == f2(16 * [1])

    def test_function_calling(self):
        f = IOH_function(1, 5, 1, suite = "BBOB")
        assert f([0] * 5) == f(np.array([0] * 5))
        assert all(f.best_so_far_variables == np.array([0] * 5))
        assert f.evaluations == 2
        assert f.best_so_far_fvalue == f(np.array([0] * 5))
        f.reset()
        assert f.evaluations == 0
        assert f.best_so_far_variables is None

    def test_random_search(self):
        def random_search(func, budget=None):
            if budget is None:
                budget = int(func.number_of_variables * 1e4)

            f_opt = np.Inf
            x_opt = None

            for _ in range(budget):
                x = np.random.uniform(func.lowerbound, func.upperbound)
                f = func(x)
                if f < f_opt:
                    f_opt = f
                    x_opt = x

            return f_opt, x_opt


        f = IOH_function(2, 5, 1)
        random_search(f)


if __name__ == '__main__':
    unittest.main()


from .utils import opt_alg, TmpdirTestCase
from IOHexperimenter import IOHexperimenter


class BenchmarkTestCase(TmpdirTestCase):    
    def test_initialization_PBO(self):
        exp = IOHexperimenter()
        exp.initialize_BBOB([1,2], [5], [5], 2, 1e-8)
        exp.set_logger_location(self.tmpdir, "run")
        a = opt_alg(1000)
        exp([a])
    
if __name__ == '__main__':
    unittest.main()


import os
import shutil
import uuid
import unittest
import numpy as np


class TmpdirTestCase(unittest.TestCase):    
    def setUp(self):
        self.tmpdir = os.path.join(os.getcwd(), str(uuid.uuid1()))

    def tearDown(self):
        shutil.rmtree(self.tmpdir)


class opt_alg:
    def __init__(self, budget):
        self.budget = budget
        self.f_opt = np.Inf
        self.x_opt = None

    def __call__(self, func):
        for _ in range(self.budget):
            x = np.random.uniform(func.lowerbound, func.upperbound)
            f = func(x)
            if f < self.f_opt:
                self.f_opt = f
                self.x_opt = x
        return self.f_opt, self.x_opt
    
    @property
    def param_rate(self):
        return np.random.randint(100)
    
    @property
    def static_param(self):
        return 123
    
    @property
    def dynamic_prop(self):
        return np.random.randint(10)


from .IOH_benchmarking import IOHexperimenter
from .IOH_function import IOH_function, custom_IOH_function, W_model_function
from .IOH_logger import IOH_logger
from .IOH_Utils import runParallelFunction, runPool, runPebblePool, runJoblib, \
    runSingleThreaded, runMPI

__all__ = [
    'IOHexperimenter', 'IOH_logger', 'IOH_function', 'custom_IOH_function', 
    'W_model_function', 'runParallelFunction', 'runPool', 'runPebblePool', 'runJoblib', 
    'runSingleThreaded', 'runMPI'
]